package de.hft.stuttgart.citydoctor2.healing;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;

import org.junit.Assert;
import org.junit.Test;

import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.CheckResult;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.ErrorId;
import de.hft.stuttgart.citydoctor2.check.ResultStatus;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.check.error.RingEdgeIntersectionError;
import de.hft.stuttgart.citydoctor2.checks.geometry.DuplicatePointsCheck;
import de.hft.stuttgart.citydoctor2.checks.geometry.RingSelfIntCheck;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.healer.Healer;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;

public class HealRingSelfIntErrorTest {
	
	@Test
	public void testIntersectingEdges1() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		Polygon p = new ConcretePolygon();
		geom.addPolygon(p);
		
		Vertex v0 = new Vertex(0, 4, 0);
		Vertex v1 = new Vertex(0, 0, 0);
		Vertex v2 = new Vertex(4, 4, 0);
		Vertex v3 = new Vertex(6, 2, 0);
		Vertex v4 = new Vertex(4, 0, 0);

		LinearRing r = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(r);
		r.addVertex(v0);
		r.addVertex(v1);
		r.addVertex(v2);
		r.addVertex(v3);
		r.addVertex(v4);
		r.addVertex(v0);
		
		geom.updateEdgesAndVertices();
		
		RingSelfIntCheck check = new RingSelfIntCheck();
		check.check(r);
		
		CheckResult cr = r.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_R_SELF_INTERSECTION, cr.getError().getErrorId());
		assertTrue(cr.getError() instanceof RingEdgeIntersectionError);

		HealRingSelfIntError healer = new HealRingSelfIntError();
		cr.getError().accept(healer, TestUtil.createDummyModificationListener());
		geom.updateEdgesAndVertices();

		assertEquals(2, geom.getPolygons().size());
		Polygon p1 = geom.getPolygons().get(0);
		Polygon p2 = geom.getPolygons().get(1);

		if (p1.getExteriorRing().getVertices().size() == 4) {
			assertEquals(5, p2.getExteriorRing().getVertices().size());
		} else {
			assertEquals(5, p1.getExteriorRing().getVertices().size());
			assertEquals(4, p2.getExteriorRing().getVertices().size());
		}
		
		check.check(p1.getExteriorRing());
		check.check(p2.getExteriorRing());
		cr = p1.getExteriorRing().getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.OK, cr.getResultStatus());
		cr = p2.getExteriorRing().getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.OK, cr.getResultStatus());
	}

	@Test
	public void testDuplicatePoint1() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		Polygon p = new ConcretePolygon();
		geom.addPolygon(p);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, 0, 0);
		Vertex v2 = new Vertex(4, 4, 0);
		Vertex v3 = new Vertex(2, 4, 0);
		Vertex v4 = new Vertex(0, 4, 0);
		
		LinearRing r = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(r);
		r.addVertex(v0);
		r.addVertex(v1);
		r.addVertex(v2);
		r.addVertex(v0);
		r.addVertex(v3);
		r.addVertex(v4);
		r.addVertex(v0);
		
		DuplicatePointsCheck check = new DuplicatePointsCheck();
		check.check(r);

		CheckResult cr = r.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_R_SELF_INTERSECTION, cr.getError().getErrorId());
		
		HealRingSelfIntError healer = new HealRingSelfIntError();
		cr.getError().accept(healer, TestUtil.createDummyModificationListener());
		
		geom.updateEdgesAndVertices();

		assertEquals(2, geom.getPolygons().size());
		Polygon p1 = geom.getPolygons().get(0);
		Polygon p2 = geom.getPolygons().get(1);
		
		assertEquals(4, p1.getExteriorRing().getVertices().size());
		assertEquals(4, p2.getExteriorRing().getVertices().size());
		
		assertEquals(2, v0.getAdjacentRings(geom).size());
	}
	
	@Test
	public void testDuplicatePoint2() {
		Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);
		Polygon p = new ConcretePolygon();
		geom.addPolygon(p);
		
		Vertex v0 = new Vertex(0, 0, 0);
		Vertex v1 = new Vertex(4, 0, 0);
		Vertex v2 = new Vertex(4, 4, 0);
		Vertex v3 = new Vertex(2, 4, 0);
		Vertex v4 = new Vertex(0, 4, 0);
		
		LinearRing r = new LinearRing(LinearRingType.EXTERIOR);
		p.setExteriorRing(r);
		r.addVertex(v4);
		r.addVertex(v0);
		r.addVertex(v1);
		r.addVertex(v2);
		r.addVertex(v0);
		r.addVertex(v3);
		r.addVertex(v4);
		
		DuplicatePointsCheck check = new DuplicatePointsCheck();
		check.check(r);

		CheckResult cr = r.getCheckResult(check);
		assertNotNull(cr);
		assertEquals(ResultStatus.ERROR, cr.getResultStatus());
		assertEquals(ErrorId.GE_R_SELF_INTERSECTION, cr.getError().getErrorId());
		
		HealRingSelfIntError healer = new HealRingSelfIntError();
		cr.getError().accept(healer, TestUtil.createDummyModificationListener());
		
		geom.updateEdgesAndVertices();

		assertEquals(2, geom.getPolygons().size());
		Polygon p1 = geom.getPolygons().get(0);
		Polygon p2 = geom.getPolygons().get(1);
		
		assertEquals(4, p1.getExteriorRing().getVertices().size());
		assertEquals(4, p2.getExteriorRing().getVertices().size());
		
		assertEquals(2, v0.getAdjacentRings(geom).size());
	}
	
	@Test
	public void testDuplicatePoint3() throws CityGmlParseException, IOException, InvalidGmlFileException {
		String path = "src/test/resources/SimpleSolid_SrefBS-GE-gml-LR-0002-T0002.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);
		Assert.assertTrue(geom.containsError(CheckId.C_GE_R_SELF_INTERSECTION));
		
		Healer healer = new Healer(c);
		healer.healCityObject(model.getBuildings().get(0));
		
		Assert.assertFalse(geom.containsError(CheckId.C_GE_R_SELF_INTERSECTION));
	}
	
	@Test
	public void testIntersectingEdges2() throws CityGmlParseException, IOException, InvalidGmlFileException {
		String path = "src/test/resources/SimpleSolid_SrefBS-GE-gml-LR-0004-T0004.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);
		Assert.assertTrue(geom.containsError(CheckId.C_GE_R_SELF_INTERSECTION));
		
		Healer healer = new Healer(c);
		healer.healCityObject(model.getBuildings().get(0));
		
		Assert.assertFalse(geom.containsError(CheckId.C_GE_R_SELF_INTERSECTION));
	}
	
	@Test
	public void testTouchingVertex() throws CityGmlParseException, IOException, InvalidGmlFileException {
		String path = "src/test/resources/SimpleSolid_SrefBS-GE-gml-LR-0004-T0005.gml";
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		config.setMinVertexDistanceInGlobalParameters(0.1);
		config.setNumberOfRoundingPlacesInGlobalParameters(8);
		CityDoctorModel model = TestUtil.loadCityModel(path, config);

		Checker c = new Checker(config, model);
		c.runChecks();

		Geometry geom = model.getBuildings().get(0).getGeometries().get(0);
		Assert.assertTrue(geom.containsError(CheckId.C_GE_R_SELF_INTERSECTION));
		
		Healer healer = new Healer(c);
		healer.healCityObject(model.getBuildings().get(0));
		
		Assert.assertFalse(geom.containsError(CheckId.C_GE_R_SELF_INTERSECTION));
	}

}
